#!/usr/bin/python3
import configparser
import json
import logging
import os
import sqlite3
import threading
import time
import apt
import apt_pkg
import dbus
import dbus.service
import netifaces
from dbus.mainloop.glib import DBusGMainLoop
from gi.repository import GLib


def singleton(class_):
    return SingletonClass(class_)

def none(*args, **kwargs):
    print("call dbus method fail")
    return

class SingletonClass:
    def __init__(self, class_):
        self.__wrapped = class_
        self.__instance = None

    def __call__(self, *args, **kwargs):
        if self.__instance is None:
            self.__instance = self.__wrapped(*args, **kwargs)
        return self.__instance


class ReturnThread(threading.Thread):
    def __init__(self, target, args=()):
        super(ReturnThread, self).__init__()
        self.func = target
        self.args = args
        self.result = None

    def run(self):
        # 接受返回值
        self.result = self.func(*self.args)

    def get_result(self):
        # 线程不结束,返回值为None
        try:
            return self.result
        except Exception as e:
            return e


def save_table(conn, table_creat, update_data, data):
    try:
        c = conn.cursor()
        c.execute(table_creat)
        conn.commit()
    except Exception as e:
        return e
    try:
        c = conn.cursor()
        c.execute(update_data, data)
        conn.commit()
    except Exception as e:
        return str(e)
    return 0


def get_network_info():
    try:
        interface = netifaces.gateways()['default'][netifaces.AF_INET][1]
        info = netifaces.ifaddresses(interface)[netifaces.AF_INET][0]['addr']
    except KeyError:
        info = "127.0.0.1"
    return info


def select_table(conn, select):
    try:
        c = conn.cursor()
        infos = c.execute(select)
        return infos
    except Exception as e:
        return str(e)


def delete_table(conn, delete, data):
    try:
        c = conn.cursor()
        c.execute(delete, data)
        conn.commit()
        return 0
    except Exception as e:
        return str(e)


def call_dbus_method(dbus_name, dbus_path, dbus_interface, def_name):
    try:
        tmp = dbus.SystemBus()
        proxy = tmp.get_object(bus_name=dbus_name, object_path=dbus_path)
        call_bus_method = dbus.Interface(proxy, dbus_interface=dbus_interface)
        way = getattr(call_bus_method, def_name)
    except Exception as e:
        logging.error("call dbus method fail for " + str(e))
        return none
    return way


def log_send(res):
    try:
        get_name = call_dbus_method('com.kylin.kydevmonit.hedronclient', '/kydevmonit/hedronclient',
                                    'com.kylin.kydevmonit.hedronclient', "get_current_user")
        hostname = get_name()
    except:
        hostname = 'root'
    if hostname == '':
        hostname = "root"
    log = {'ip': get_network_info(), 'time': time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()),
           'hostname': hostname, 'name': "apt-install", 'lv': "", 'message': ""}
    if res["Success"] == True:
        lv = "info"
        message = json.dumps(res["Change"])
    elif res["Success"] == "warning":
        lv = "warning"
        message = res["Error"]
    else:
        lv = "error"
        message = res["Error"]
    message += " uuid " + res["uuid"]
    log['lv'] = lv
    log['message'] = message
    logging.debug(log)
    logsend = call_dbus_method('org.log.sys_transmit', '/org/log/sys_transmit', 'org.log.transmit', "log_transmit")
    logsend(json.dumps(log))


@singleton
class AptInUp(dbus.service.Object):
    """
        add_process     访问接口，将安装任务写入数据库
        star_install    读取并任务的线程
        process_info    收集安装进度信息
        install_update      安装执行主体
        get_version     获取客户端版本信息
        get_confirm     用户拒绝的前端调取函数
    """

    def __init__(self):
        self.dbusMainLoop = DBusGMainLoop(set_as_default=True)
        self.loop = GLib.MainLoop()
        self.bus = dbus.SystemBus()
        bus_name = dbus.service.BusName('com.kylin.kydevmonit.apt_software', bus=self.bus)
        dbus.service.Object.__init__(self, bus_name, '/com/kylin/kydevmonit/apt_software')
        self.cache = object
        self.bus.add_signal_receiver(handler_function=self.add_process, signal_name="install_signal",
                                     bus_name="com.kylin.kydevmonit.hedronsingle",
                                     path="/com/kylin/kydevmonit/hedron_single")
        self.bus.add_signal_receiver(handler_function=self.get_confirm, signal_name="update_result_signal",
                                     bus_name="com.kylin.kydevmonit.hedronagent",
                                     path="/kydevmonit/Responser")
        self.update_progress = apt.progress.base.AcquireProgress()
        self.install_progress = apt.progress.base.InstallProgress()
        self.system_flag = "ready"
        self.system_install = False
        self.system_lock = False
        self.config_reboot = False
        self.config_refuse_open = False
        self.config_silent = False
        self.congif_timeout = True
        self.uninstall = False
        self.conn = sqlite3.connect("/var/log/hedron/db.sqlite", check_same_thread=False)
        self.conn.cursor().execute("CREATE TABLE IF NOT EXISTS REMAIN(id integer PRIMARY KEY autoincrement,NAME TEXT,UUID TEXT)")
        self.conn.commit()
        self.confrim = ""
        self.__instantiate()
        info = threading.Thread(target=self.star_install, args=[])
        info.start()
        self.__check_install()

    def __instantiate(self):
        """
            配置初始化函数
        """
        try:
            config_path = os.path.join("/etc/hedron/apt-install.ini")
            cf = configparser.ConfigParser()
            cf.read(config_path)
            self.config_refuse_open = cf.get("apt-install", "refuse_open")
            self.config_silent = cf.get("apt-install", "refuse_silent")
            self.congif_timeout = cf.get("apt-install", "refuse_timeout")
        except Exception as e:
            pass

    def __check_install(self):
        """
            安装检查函数
        """
        result = select_table(self.conn, "SELECT NAME, UUID from REMAIN")
        res = result.fetchall()
        if res:
            self.system_install = True
        else:
            self.system_install = False
        return res

    @dbus.service.method('com.kylin.kydevmonit.apt_software', in_signature='s')
    def add_process(self, message):
        message = json.loads(message)
        save_table(self.conn,
                   "CREATE TABLE IF NOT EXISTS REMAIN(id integer PRIMARY KEY autoincrement,NAME TEXT,UUID TEXT)",
                   "INSERT INTO REMAIN(Name, Uuid) VALUES (?, ?)", [message["pkg"], message["uuid"]])
        self.system_install = True
        return 0

    @dbus.service.method('com.kylin.kydevmonit.apt_software')
    def get_version(self):
        try:
            hedron_package = list(filter(lambda pkg: pkg.name == 'hedron-client', apt.Cache()))[0]
            version = hedron_package.installed.version
        except Exception:
            version = 'NULL'
        return version

    @dbus.service.signal(dbus_interface='com.kylin.kydevmonit.apt_software', signature='s')
    def result_signal(self, info):
        return info

    def star_install(self):
        while True:
            if self.system_install:
                res = self.__check_install()
                for tup in res:
                    if not self.system_lock:
                        try:
                            if self.config_refuse_open:
                                obj = call_dbus_method("com.kylin.kydevmonit.hedronagent", "/kydevmonit/Responser",
                                                       "com.kylin.kydevmonit.hedronagent", "update_request")
                                obj(tup[0])
                        except:
                            pass
                        self.install_update(tup[0], tup[1])
            else:
                time.sleep(5)

    def install_update(self, message, uuid):
        # 加锁
        self.system_lock = True
        # 字符串化列表
        package = message.split(',')
        # front为前端状态，获取接口失败时静默安装
        front = True
        # lock为dpkg是否加锁，flag为等待次数上线
        lock = True
        flag = 0
        while lock:
            try:
                apt_pkg.pkgsystem_lock()
                apt_pkg.pkgsystem_unlock()
                lock = False
            except Exception:
                flag += 1
                # 等待间隔
                time.sleep(30)
            # 超时处理
            if flag >= 20:
                result = {"Success": False, "Error": "Error: time out while try get lock", "Change": {}, "uuid": uuid}
                self.system_lock = False
                log_send(result)
                return -1
        try:
            # 前端接口调取
            obj = call_dbus_method("com.kylin.kydevmonit.hedronagent", "/kydevmonit/Responser",
                                   "com.kylin.kydevmonit.hedronagent", "update")
        except Exception:
            front = False
        # 安装主体执行
        confrim = self.__confirm_install(uuid, package, front)
        self.confrim = ""
        if confrim:
            act = ReturnThread(target=self.__install_packages, args=[package, uuid, False])
            act.start()
            # 前端进度交互
            while self.system_flag != "finished" and self.system_flag != "error" and self.system_flag != "ready" and front:
                try:
                    info = self.process_info()
                    obj(info)
                except Exception:
                    logging.error("agnat lost")
                time.sleep(0.2)
            act.join()
            if front:
                try:
                    info = self.process_info(self.uninstall)
                    obj(info)
                except Exception:
                    logging.error("agnat lost")
            self.system_flag = "ready"  
            log_send(act.get_result())
            self.result_signal('{"type":"upgrade_complete"}')
        # 删除执行过的命令
        delete_table(self.conn, "DELETE from REMAIN where UUID = ?", [uuid])
        self.__check_install()
        self.system_lock = False
        if self.config_reboot:
            try:
                com = call_dbus_method("com.kylin.kydevmonit.hedronagent", "/kydevmonit/Responser",
                                       "com.kylin.kydevmonit.hedronagent", "rebootORshutdown")
                com(False, '{"type": "apt", "message": "", "status": 0, "uuid": ""}', "安全管控客户端更新完成，是否重启")
            except Exception:
                logging.error("front silent")
            self.config_reboot = False
        return 0

    def get_confirm(self, accept):
        self.confrim = accept

    def __confirm_install(self, uuid, package, front):
        """
            拒绝确认函数
            front 为所需要的对接的前端，不存在为false
        """
        if self.config_refuse_open:
            if not front:
                return self.config_silent
            t = 0
            while t <= 30:
                time.sleep(1)
                t = t + 1
                if self.confrim == "True":
                    return True
                elif self.confrim == "False":
                    result = {"Success": "warning", "Error": "终端用户拒绝了{}安装".format(package), "Change": {}, "uuid": uuid}
                    log_send(result)
                    return False
            return self.congif_timeout
        else:
            return True

    def process_info(self, uninstall= False):
        info = {'current_cps': 0,
                'value': 0,
                'status': self.system_flag,
                'message': ""}
        if self.system_flag == "update":
            if self.update_progress.total_items == 0:
                val = 0
            else:
                val = (self.update_progress.current_items / self.update_progress.total_items) * 100
            info['current_cps'] = self.update_progress.current_cps
            info['value'] = val
            info['status'] = self.system_flag
        elif self.system_flag == "commit":
            info['value'] = self.install_progress.percent
        if uninstall:
            info['status'] = self.system_flag
            info['message'] = "部分软件已是最新版，未执行安装"
        return json.dumps(info)

    def __install_packages(self, packages, uuid, dist_upgrade=False):
        """
            apt安装主体
            packages为安装列表
        """
        # 源更新
        result = {"Success": True, "Error": "", "Change": {}, "uuid": uuid}
        self.system_flag = "update"
        try:
            self.cache = apt.Cache()
            self.cache.update()
            self.cache.open()
        except Exception as e:
            self.system_flag = "error"
            result['Success'] = False
            result['Error'] = "源更新失败：" + str(e)
            result['Change'] = []
            return result
        self.system_flag = "commit"
        # 安装流程
        try:
            for package_name in packages:
                pkg = self.cache[package_name]
                if pkg.installed is not None:
                    pkg.mark_upgrade()
                else:
                    pkg.mark_install()
            if dist_upgrade:
                self.cache.upgrade(True)
            self.cache.commit(install_progress=self.install_progress)
            # 安装结束，生成结果
            self.system_flag = "finished"
            error = None
            changed_packages = self.cache.get_changes()
            new_cache = apt.Cache()
            changes = []
            fail_pkg = ""
            try:
                for package in changed_packages:
                    fail_pkg = package.name
                    name = package.name
                    original_version = package.installed.version if package.installed is not None else "None"
                    new_version = new_cache[name].installed.version
                    if name == "hedron-client" or name == "kim-client":
                        self.config_reboot = True
                    changes.append({
                        'name': name,
                        'original_version': original_version,
                        'new_version': new_version
                    })
                installed_set = set([item['name'] for item in changes])
                original_set = set(packages)
                unupgraded_set = original_set - installed_set
                for item in unupgraded_set:
                    fail_pkg = item
                    version = new_cache[item].installed.version
                    changes.append({
                        'name': item,
                        'original_version': version,
                        'new_version': version
                    })
                    self.uninstall = True
            except Exception as e:
                result['Success'] = False
                error = "{} 安装失败，请检查依赖等是否完全".format(fail_pkg)
                changes = []
                self.system_flag = "error"
        except Exception as e:
            result['Success'] = False
            error = "包下载失败：" + str(e)
            changes = []
            self.system_flag = "error"
        result['Error'] = error
        result['Change'] = changes
        return result


if __name__ == "__main__":
    begin = AptInUp()
    begin.loop.run()
